"""Unit tests for audio file handling module."""

from pathlib import Path
from unittest.mock import Mock, patch

import pytest
from mutagen.oggopus import OggOpus
from soundfile import SoundFile

from epub2audio.audio_handler import AudioHandler
from epub2audio.config import ErrorCodes
from epub2audio.epub_processor import BookMetadata
from epub2audio.helpers import AudioHandlerError


@pytest.fixture
def sample_metadata() -> BookMetadata:
    """Create sample book metadata."""
    return BookMetadata(
        title="Test Book",
        creator="Test Author",
        date="2025",
        identifier="id123",
        language="en",
        publisher="Test Publisher",
        description="Test Description",
    )


@pytest.fixture
def audio_handler(tmp_path: Path, sample_metadata: BookMetadata) -> AudioHandler:
    """Create an AudioHandler instance."""
    epub_path = str(tmp_path / "test.epub")
    output_path = str(tmp_path / "test.ogg")
    # Create an empty epub file for testing
    with open(epub_path, "w") as f:
        f.write("dummy content")
    return AudioHandler(epub_path, output_path, sample_metadata, quiet=True)


def test_audio_handler_init(
    audio_handler: AudioHandler, sample_metadata: BookMetadata
) -> None:
    """Test AudioHandler initialization."""
    assert audio_handler is not None
    assert audio_handler.metadata == sample_metadata
    assert audio_handler.chapter_markers == []
    assert audio_handler.total_chapters == 0
    assert audio_handler.total_duration == 0.0


def test_add_chapter_marker(audio_handler: AudioHandler) -> None:
    """Test adding chapter markers."""
    audio_handler.add_chapter_marker("Chapter 1", 0.0, 10.0)
    audio_handler.add_chapter_marker("Chapter 2", 10.0, 20.0)

    assert len(audio_handler.chapter_markers) == 2

    # Check first chapter marker
    first_chapter = audio_handler.chapter_markers[0]
    assert first_chapter.title == "Chapter 1"
    assert first_chapter.start_time == 0.0
    assert first_chapter.end_time == 10.0
    assert first_chapter.start_time_str == "00:00:00.000"
    assert first_chapter.end_time_str == "00:00:10.000"

    # Check second chapter marker
    second_chapter = audio_handler.chapter_markers[1]
    assert second_chapter.title == "Chapter 2"
    assert second_chapter.start_time == 10.0
    assert second_chapter.end_time == 20.0
    assert second_chapter.start_time_str == "00:00:10.000"
    assert second_chapter.end_time_str == "00:00:20.000"


def test_write_metadata(
    audio_handler: AudioHandler, sample_metadata: BookMetadata
) -> None:
    """Test writing metadata to audio file."""
    mock_audio_file = Mock(spec=OggOpus)
    mock_audio_file.__setitem__ = Mock()

    audio_handler.add_chapter_marker("Chapter 1", 0.0, 10.0)
    with patch.object(audio_handler, "_write_vorbis_metadata") as mock_write_vorbis:
        audio_handler._write_metadata(mock_audio_file)
        mock_write_vorbis.assert_called_once_with(mock_audio_file)


def test_finalize_audio_file(audio_handler: AudioHandler, tmp_path: Path) -> None:
    """Test finalizing audio file."""
    # Create a mock audio segment
    mock_segment = Mock(spec=SoundFile)
    mock_segment.name = str(tmp_path / "test_segment.ogg")
    mock_segment.close = Mock()

    # Create mock for concatenate_segments
    with patch.object(audio_handler, "_concatenate_segments") as mock_concatenate:
        mock_concatenate.return_value = mock_segment

        with patch("epub2audio.audio_handler.SUPPORTED_AUDIO_FORMATS") as mock_formats:
            # Mock the types
            mock_type = type("MockOggOpus", (), {})
            format_info_mock = Mock()
            format_info_mock.file_class = mock_type
            mock_formats.__getitem__.return_value = format_info_mock

            # Mock the type initialization
            mock_file_instance = Mock()
            with patch.object(mock_type, "__new__", return_value=mock_file_instance):
                # Mock the file operations
                with patch.object(
                    audio_handler, "_write_metadata"
                ) as mock_write_metadata:
                    # Mock shutil.move to prevent file operation
                    with patch("epub2audio.audio_handler.move") as mock_move:
                        audio_handler.finalize_audio_file([mock_segment])

                        # Check if metadata was written
                        mock_formats.__getitem__.assert_called_with(
                            audio_handler.extension
                        )
                        mock_write_metadata.assert_called_once_with(mock_file_instance)
                        mock_file_instance.save.assert_called_once()
                        # Check file was moved
                        mock_move.assert_called_once_with(
                            mock_segment.name, audio_handler.output_path
                        )


def test_finalize_audio_file_error(audio_handler: AudioHandler, tmp_path: Path) -> None:
    """Test error handling in finalize_audio_file."""
    # Create a mock audio segment
    mock_segment = Mock(spec=SoundFile)
    mock_segment.name = str(tmp_path / "test_segment.ogg")
    mock_segment.close = Mock()

    # Create mock for concatenate_segments
    with patch.object(audio_handler, "_concatenate_segments") as mock_concatenate:
        mock_concatenate.return_value = mock_segment

        with patch("epub2audio.audio_handler.SUPPORTED_AUDIO_FORMATS") as mock_formats:
            # Set up mock to raise an exception
            file_class_mock = Mock(side_effect=Exception("Save error"))
            format_info_mock = Mock()
            format_info_mock.file_class = file_class_mock
            mock_formats.__getitem__.return_value = format_info_mock

            with pytest.raises(AudioHandlerError) as exc_info:
                audio_handler.finalize_audio_file([mock_segment])
            assert exc_info.value.error_code == ErrorCodes.FILESYSTEM_ERROR


def test_total_chapters(audio_handler: AudioHandler) -> None:
    """Test total chapters property."""
    assert audio_handler.total_chapters == 0

    audio_handler.add_chapter_marker("Chapter 1", 0.0, 10.0)
    assert audio_handler.total_chapters == 1

    audio_handler.add_chapter_marker("Chapter 2", 10.0, 20.0)
    assert audio_handler.total_chapters == 2


def test_total_duration(audio_handler: AudioHandler) -> None:
    """Test total duration property."""
    assert audio_handler.total_duration == 0.0

    audio_handler.add_chapter_marker("Chapter 1", 0.0, 10.0)
    assert audio_handler.total_duration == 10.0

    audio_handler.add_chapter_marker("Chapter 2", 10.0, 20.0)
    assert audio_handler.total_duration == 20.0



"""Audio file handling and metadata management."""

import base64
import io
from dataclasses import dataclass
from pathlib import Path
from shutil import move

import mutagen
from loguru import logger
from mutagen.flac import FLAC, Picture
from mutagen.id3 import (
    CHAP,
    COMM,
    CTOC,
    TCOP,
    TDRC,
    TIT2,
    TOR,
    TPE1,
    TPE2,
    TPUB,
    CTOCFlags,
    PictureType,
)
from mutagen.mp3 import MP3
from mutagen.oggflac import OggFLAC
from mutagen.oggopus import OggOpus
from PIL import Image
from soundfile import SoundFile
from tqdm import tqdm  # type: ignore

from .config import SUPPORTED_AUDIO_FORMATS, ErrorCodes
from .epub_processor import BookMetadata
from .helpers import AudioHandlerError, CacheDirManager, StrPath, format_time


@dataclass
class ChapterMarker:
    """Class for storing chapter marker information."""

    title: str
    start_time: float
    end_time: float

    @property
    def start_time_str(self) -> str:
        """Get the start time as a string."""
        return format_time(self.start_time)

    @property
    def end_time_str(self) -> str:
        """Get the end time as a string."""
        return format_time(self.end_time)

    @property
    def duration(self) -> float:
        """Get the duration of the chapter."""
        return self.end_time - self.start_time


class AudioHandler:
    """Class for handling audio file creation and metadata."""

    def __init__(
        self,
        epub_path: StrPath,
        output_path: StrPath,
        metadata: BookMetadata,
        quiet: bool = True,
    ):
        """Initialize the audio handler.

        Args:
            epub_path: Path to the EPUB file
            output_path: Path to the output audio file
            metadata: Book metadata
            quiet: Whether to suppress progress bars
        """
        self.epub_path = Path(epub_path)
        self.output_path = Path(output_path)
        self.extension = self.output_path.suffix
        self.metadata = metadata
        self.cache_dir_manager = CacheDirManager(epub_path, extension=self.extension)
        self.chapter_markers: list[ChapterMarker] = []
        self.quiet = quiet

    def add_chapter_marker(
        self, title: str, start_time: float, end_time: float
    ) -> None:
        """Add a chapter marker.

        Args:
            title: Chapter title
            start_time: Start time in seconds
            end_time: End time in seconds
        """
        self.chapter_markers.append(ChapterMarker(title, start_time, end_time))

    def _make_flac_picture(self) -> Picture:
        """Make a FLAC picture.

        Returns:
            Picture: FLAC picture
        """
        if not self.metadata.cover_image:
            raise ValueError("No cover image found")
        cover_image_bytes = base64.b64decode(self.metadata.cover_image)
        cover_image = Image.open(io.BytesIO(cover_image_bytes))

        # cover_image.show()
        cover_picture = Picture()
        cover_picture.data = cover_image_bytes  # cover_image.tobytes()
        if cover_image.format:
            cover_picture.mime = f"image/{cover_image.format.lower()}"
        else:
            cover_picture.mime = "image/jpeg"
        cover_picture.type = PictureType.COVER_FRONT
        cover_picture.height = cover_image.height
        cover_picture.width = cover_image.width
        cover_picture.depth = 8
        cover_picture.colors = 0
        cover_picture.desc = "Cover image"  # TODO: Add description
        return cover_picture

    def _parse_cover_image(self) -> list[str]:
        """Parse the cover image.

        Args:
            cover_image_str: Cover image as a base64 encoded string
        """
        cover_picture = self._make_flac_picture()
        cover_b64_str = base64.b64encode(cover_picture.write()).decode("ascii")
        # cover_image.thumbnail((256, 256))
        # cover_picture.data = cover_image.tobytes()
        cover_picture.type = PictureType.FILE_ICON
        file_b64_str = base64.b64encode(cover_picture.write()).decode("ascii")
        return [
            cover_b64_str,
            file_b64_str,
        ]

    def _write_mp3_metadata(self, audio_file: MP3) -> None:
        """Write metadata to the audio file.

        Args:
            audio_file: MP3 file object
        """
        # Create ID3 tag if it doesn't exist
        if audio_file.tags is None:
            audio_file.add_tags()
            assert audio_file.tags is not None

        # Add basic metadata
        audio_file.tags.add(TIT2(encoding=3, text=[self.metadata.title]))
        if self.metadata.creator:
            audio_file.tags.add(TPE1(encoding=3, text=[self.metadata.creator]))
        if self.metadata.date:
            audio_file.tags.add(TDRC(encoding=3, text=[self.metadata.date]))
        if self.metadata.publisher:
            audio_file.tags.add(TPUB(encoding=3, text=[self.metadata.publisher]))
        if self.metadata.description:
            audio_file.tags.add(COMM(encoding=3, text=[self.metadata.description]))

        # Add chapter markers
        if self.chapter_markers:
            logger.debug(f"Adding chapter markers to {self.output_path}")
            # Create chapter IDs
            chapter_ids = [f"chp{i:03d}" for i in range(len(self.chapter_markers))]

            # Create CTOC (Table of Contents)
            audio_file.tags.add(
                CTOC(
                    element_id="toc",
                    flags=CTOCFlags.TOP_LEVEL | CTOCFlags.ORDERED,
                    child_element_ids=chapter_ids,
                    sub_frames=[TIT2(encoding=3, text=["Table of Contents"])],
                )
            )

            # Add individual chapters
            for i, marker in enumerate(self.chapter_markers):
                # Convert times to milliseconds
                start_time = int(marker.start_time * 1000)
                end_time = int(marker.end_time * 1000)
                logger.trace(f"Adding chapter '{marker.title}' at {start_time}ms")
                audio_file.tags.add(
                    CHAP(
                        element_id=chapter_ids[i],
                        start_time=start_time,
                        end_time=end_time,
                        sub_frames=[TIT2(encoding=3, text=[marker.title])],
                    )
                )

        # Add organization info
        audio_file.tags.add(TOR(encoding=3, text=["epub2audio"]))
        audio_file.tags.add(TPE2(encoding=3, text=["Kokoro TextToSpeech"]))
        audio_file.tags.add(
            TCOP(encoding=3, text=["https://creativecommons.org/licenses/by-sa/4.0/"])
        )

    def _write_vorbis_metadata(self, audio_file: OggOpus | OggFLAC | FLAC) -> None:
        """Write metadata to the audio file.

        Args:
            audio_file: mutagen.FileType file object that uses Vorbis Comments
        """
        # Add basic metadata
        audio_file["TITLE"] = self.metadata.title
        if self.metadata.creator:
            audio_file["ARTIST"] = self.metadata.creator
        if self.metadata.date:
            audio_file["DATE"] = self.metadata.date
        if self.metadata.publisher:
            audio_file["PUBLISHER"] = self.metadata.publisher
        if self.metadata.description:
            audio_file["DESCRIPTION"] = self.metadata.description
        if self.metadata.cover_image:
            if isinstance(audio_file, FLAC):
                flac_picture = self._make_flac_picture()
                audio_file.add_picture(flac_picture)
                flac_picture.type = PictureType.FILE_ICON
                audio_file.add_picture(flac_picture)
            else:
                audio_file["METADATA_BLOCK_PICTURE"] = self._parse_cover_image()

        audio_file["ORGANIZATION"] = "epub2audio"
        audio_file["PERFORMER"] = "Kokoro TextToSpeech"
        audio_file["COPYRIGHT"] = "https://creativecommons.org/licenses/by-sa/4.0/"

        # Add chapter markers
        for i, marker in enumerate(self.chapter_markers):
            audio_file[f"CHAPTER{i:03d}NAME"] = marker.title
            audio_file[f"CHAPTER{i:03d}"] = marker.start_time_str

    def _write_metadata(self, audio_file: mutagen.FileType) -> None:
        """Write metadata to the audio file.

        Args:
            audio_file: mutagen.FileType file object to write metadata to
        """
        if isinstance(audio_file, OggOpus):
            self._write_vorbis_metadata(audio_file)
        elif isinstance(audio_file, OggFLAC):
            self._write_vorbis_metadata(audio_file)
        elif isinstance(audio_file, FLAC):
            self._write_vorbis_metadata(audio_file)
        elif isinstance(audio_file, MP3):
            self._write_mp3_metadata(audio_file)
        else:
            raise ValueError(f"Unsupported audio file type: {type(audio_file)}")

    def _concatenate_segments(self, segments: list[SoundFile]) -> SoundFile:
        """Concatenate multiple audio segments.

        Args:
            segments: List of audio segments to concatenate

        Returns:
            SoundFile: Concatenated audio
        """
        if not segments:
            raise ValueError("No audio segments to concatenate")

        # Ensure all segments have the same sample rate
        sample_rate = segments[0].samplerate
        if not all(s.samplerate == sample_rate for s in segments):
            raise ValueError("All audio segments must have the same sample rate")

        # Concatenate the audio data
        temp_file = self.cache_dir_manager.get_file("concatenated")
        format_info = SUPPORTED_AUDIO_FORMATS[self.extension]
        concatenated_data = SoundFile(
            temp_file,
            mode="w",
            samplerate=sample_rate,
            channels=1,
            format=format_info.format,
            subtype=format_info.subtype,
        )
        with tqdm(
            total=sum(segment.frames for segment in segments),
            desc="Concatenating audio segments",
            disable=self.quiet,
            unit="frames",
        ) as pbar:
            for segment in segments:
                with SoundFile(segment.name, mode="r") as sf:
                    data = sf.read()
                concatenated_data.write(data)
                pbar.update(len(data))
        concatenated_data.close()
        return SoundFile(concatenated_data.name)

    def finalize_audio_file(self, segments: list[SoundFile]) -> None:
        """Write the final audio file with metadata.

        Args:
            segments: List of audio segments to concatenate and write to the final file

        Raises:
            AudioHandlerError: If writing the audio file fails
        """
        final_segment = self._concatenate_segments(segments)
        try:
            # Add metadata
            logger.trace(f"Adding metadata to final audio file, {final_segment.name}")
            final_segment.close()
            audio_file_class = SUPPORTED_AUDIO_FORMATS[self.extension].file_class
            audio_file = audio_file_class(final_segment.name)
            self._write_metadata(audio_file)
            logger.trace(f"Saving final audio file {audio_file.pprint()}")
            audio_file.save()
            move(final_segment.name, self.output_path)
            logger.debug(f"Final audio file saved to {self.output_path}")

        except Exception as e:
            raise AudioHandlerError(
                f"Failed to write final audio file: {str(e)}",
                ErrorCodes.FILESYSTEM_ERROR,
            ) from e

    @property
    def total_chapters(self) -> int:
        """Get the total number of chapters.

        Returns:
            int: Number of chapters
        """
        return len(self.chapter_markers)

    @property
    def total_duration(self) -> float:
        """Get the total duration in seconds.

        Returns:
            float: Total duration
        """
        if not self.chapter_markers:
            return 0.0
        return self.chapter_markers[-1].end_time



"""EPUB processing and text extraction module."""

import base64
import re
from dataclasses import dataclass
from typing import Optional

import ebooklib  # type: ignore
from bs4 import BeautifulSoup
from ebooklib import epub
from loguru import logger

from .config import METADATA_FIELDS, ErrorCodes, WarningTypes
from .helpers import ConversionError, ConversionWarning, StrPath


@dataclass
class Chapter:
    """Class representing a chapter in the EPUB book."""

    title: str
    content: str
    order: int
    id: str

    def __str__(self) -> str:
        """Return a string representation of the chapter."""
        return (
            f"Chapter(title='{self.title}', id='{self.id}', order={self.order}, "
            + f"content='{self.content[: len(self.title)]}...')"
        )

    def __eq__(self, other: object) -> bool:
        """Check if two chapters are equal."""
        if not isinstance(other, Chapter):
            return False
        return (
            self.id == other.id
            and self.order == other.order
            and self.title == other.title
            and self.content == other.content
        )

    def __lt__(self, other: object) -> bool:
        """Compare two chapters."""
        if not isinstance(other, Chapter):
            return False
        return self.order < other.order

    def __le__(self, other: object) -> bool:
        """Compare two chapters."""
        if not isinstance(other, Chapter):
            return False
        return self.order <= other.order


@dataclass
class BookMetadata:
    """Class representing book metadata."""

    title: str
    creator: Optional[str] = None
    date: Optional[str] = None
    identifier: Optional[str] = None
    language: Optional[str] = None
    publisher: Optional[str] = None
    description: Optional[str] = None
    cover_image: Optional[str] = None  # Base64 encoded str

    @property
    def book_sentence(self) -> str:
        """Get a sentence that describes the book."""
        return (
            f"A book by {self.creator}, published in {self.date} by {self.publisher}."
            + "Audio generated by Epub2Audio."
        )

    def __str__(self) -> str:
        """Return a string representation of the book metadata."""
        return (
            f"BookMetadata(title={self.title}, "
            + (f"creator={self.creator}, " if self.creator else "")
            + (f"date={self.date}, " if self.date else "")
            + (f"publisher={self.publisher}, " if self.publisher else "")
            + (f"description={self.description}, " if self.description else "")
            + (f"cover_image={len(self.cover_image)}, " if self.cover_image else "")
            + ")"
        )


@dataclass
class CoverImage:
    """Class representing a cover image."""

    image: bytes
    width: int
    height: int


@dataclass
class Book:
    """Class representing a book."""

    metadata: BookMetadata
    chapters: list[Chapter]

    def __str__(self) -> str:
        """Return a string representation of the book."""
        return (
            f"Book(\n\tmetadata={self.metadata},\n\t"
            + f"chapters={'\n\t\t'.join(str(chapter) for chapter in self.chapters)}\n)"
        )


def get_book_length(chapters: list[Chapter]) -> int:
    """Get the length of an EPUB book.

    Args:
        chapters: List of chapters
    """
    return sum(len(chapter.content) for chapter in chapters)


class EpubProcessor:
    """Class for processing EPUB files and extracting content."""

    def __init__(self, epub_path: StrPath):
        """Initialize the EPUB processor.

        Args:
            epub_path: Path to the EPUB file

        Raises:
            ConversionError: If the EPUB file is invalid or cannot be read
        """
        self.epub_path = epub_path
        self.warnings: list[ConversionWarning] = []
        try:
            self.epub = epub.read_epub(epub_path, options={"ignore_ncx": True})
        except Exception as e:
            raise ConversionError(
                f"Failed to read EPUB file: {str(e)}", ErrorCodes.INVALID_EPUB
            ) from e

        self._chapters_in_one_page = False
        self.metadata = self._extract_metadata()
        self.chapters = self._extract_chapters()
        if len(self.chapters) == 0:
            raise ConversionError(
                "No valid chapters found in EPUB file", ErrorCodes.INVALID_EPUB
            )
        if len(self.chapters) == 1:
            logger.warning("Only one chapter found in EPUB file")
            self.warnings.append(
                ConversionWarning(
                    type=WarningTypes.FORMATTING_ISSUE,
                    message="Only one chapter found in EPUB file",
                )
            )
        self.book = Book(metadata=self.metadata, chapters=self.chapters)

        logger.debug(f"Metadata: {self.metadata}")
        logger.trace(f"Number of chapters: {len(self.chapters)}")
        logger.trace(f"Chapter titles: {[chapter.title for chapter in self.chapters]}")
        logger.trace(f"Book length: {get_book_length(self.chapters)}")

    def _extract_metadata(self) -> BookMetadata:
        """Extract metadata from the EPUB file.

        Returns:
            BookMetadata: Extracted metadata
        """
        metadata = {}

        # Extract Dublin Core metadata
        for field in METADATA_FIELDS:
            value = self.epub.get_metadata("DC", field)
            if value:
                metadata[field] = value[0][0]
            else:
                if field == "title":
                    raise ConversionError(
                        "EPUB file missing required title metadata",
                        ErrorCodes.INVALID_EPUB,
                    )
                logger.warning(f"Missing metadata field: {field}")
                self.warnings.append(
                    ConversionWarning(
                        type=WarningTypes.UNSUPPORTED_METADATA,
                        message=f"Missing metadata field: {field}",
                    )
                )

        for cover in self.epub.get_items_of_type(ebooklib.ITEM_COVER):
            logger.trace(f"Cover: {cover}")
            cover_bytes = cover.get_content()
            # cover_image = Image.open(io.BytesIO(cover_bytes))
            # cover_image.thumbnail((256, 256))
            # cover_image.show()
            # TODO: if more than one cover,
            # show the cover images to the user, or use the largest
            if cover_bytes:
                metadata["cover_image"] = base64.b64encode(cover_bytes).decode("utf-8")
                # just take the first cover
                break

        return BookMetadata(**metadata)

    def _clean_text(self, html_content: str) -> str:
        """Clean HTML content and extract plain text.

        Args:
            html_content: Raw HTML content

        Returns:
            str: Cleaned text content
        """
        soup = BeautifulSoup(html_content, "html.parser")

        # Remove script and style elements
        for element in soup(["script", "style"]):
            element.decompose()

        # Handle non-text elements
        for element in soup.find_all(["img", "svg"]):
            logger.warning(f"Skipping non-text element: {element.name}")
            self.warnings.append(
                ConversionWarning(
                    type=WarningTypes.NON_TEXT_ELEMENT,
                    message=f"Skipping non-text element: {element.name}",
                )
            )
            element.decompose()

        # Get text and clean it
        text = soup.get_text()

        # Normalize whitespace
        text = re.sub(r"\s+", " ", text)
        text = text.strip()

        return text

    def _flatten_table_of_contents(
        self, table_of_contents: list[epub.Section | epub.Link]
    ) -> list[epub.EpubItem | epub.Link]:
        """Flatten the table of contents.

        Args:
            table_of_contents: Table of contents

        Returns:
            list[epub.EpubItem]: Flattened list of EPUB items
        """
        logger.trace(f"length of toc: {len(table_of_contents)}")
        flattened_table_of_contents = []
        for item in table_of_contents:
            if isinstance(item, epub.Link):
                link_split = item.href.split("#")
                if len(link_split) > 1:
                    # link within one big page
                    big_page = self.epub.get_item_with_href(link_split[0])
                    if (
                        big_page not in flattened_table_of_contents
                        and not self._chapters_in_one_page
                    ):
                        flattened_table_of_contents.append(big_page)
                    else:
                        self._chapters_in_one_page = True
                    flattened_table_of_contents.append(item)
                else:
                    item_from_link = self.epub.get_item_with_href(item.href)
                    if (
                        item_from_link in flattened_table_of_contents
                        or item_from_link in table_of_contents
                    ):
                        self._chapters_in_one_page = True
                    flattened_table_of_contents.append(item_from_link)
            elif isinstance(item, epub.Section):
                flattened_table_of_contents.extend(
                    self._flatten_table_of_contents(item.children)
                )
            elif isinstance(item, tuple):
                section, children = item
                flattened_table_of_contents.extend(
                    self._flatten_table_of_contents(children)
                )
            else:
                logger.warning(f"Skipping non-link or section item: {item}")
                self.warnings.append(
                    ConversionWarning(
                        type=WarningTypes.FORMATTING_ISSUE,
                        message=f"Skipping non-link or section item: {item}",
                    )
                )
        return flattened_table_of_contents

    def _get_chapters_in_one_page(
        self, item: epub.EpubItem, table_of_contents: list[epub.Link]
    ) -> list[Chapter]:
        """Get the chapters from a single page EPUB format.

        Args:
            item: The HTML content of the book page
            table_of_contents: The table of contents of the book

        Returns:
            list[Chapter]: List of chapters extracted from the page
        """
        chapters: list[Chapter] = []
        order = 0

        raw_content = item.get_content().decode("utf-8")
        soup = BeautifulSoup(raw_content, "html.parser")

        # Add book title if it exists
        title_tag = soup.find("p", class_="Title")
        if title_tag:
            title = title_tag.get_text(strip=True)
            if title.lower() != self.metadata.title.lower():
                logger.trace(
                    f"Book title found: {title}, "
                    f"not metadata title: {self.metadata.title}"
                )
                book_sentence = f"{title}. Audio generated by Epub2Audio."
                chapters.insert(
                    0, Chapter(title=title, content=book_sentence, order=-1, id="title")
                )

        # Find all chapter headings (they have class "Heading-2")
        chapter_heads = soup.find_all("p", class_="Heading-2")

        for head in chapter_heads:
            # Get the chapter title
            title = head.get_text(strip=True)
            # if not title.startswith("Chapter"):
            #     continue

            # Get the chapter content by finding all content between this heading
            # and the next heading (or end of document)
            content = []
            current = head.find_next_sibling()
            while current and not (
                current.name
                and current.name == "p"
                and "Heading-2" in current.get("class", [])
            ):
                if current.name and current.name == "p":
                    content.append(current.get_text(strip=True))
                current = current.find_next_sibling()

            # Join the content paragraphs
            chapter_content = " ".join(content)
            order += 1

            # Create the chapter
            chapter = Chapter(
                title=title,
                content=chapter_content,
                order=order,
                id=head.get("id", f"chapter_{order}"),
            )
            chapters.append(chapter)

        return chapters

    def _get_title_chapter(self) -> Chapter:
        """Get the title chapter."""
        return Chapter(
            title=self.metadata.title,
            content=self.metadata.book_sentence,
            order=-1,
            id="title",
        )

    def _get_chapters_from_pages(
        self,
        table_of_contents: list[epub.EpubItem],
    ) -> list[Chapter]:
        """Get the chapters from the pages of the EPUB file.

        Args:
            table_of_contents: The table of contents of the book
        """
        chapters: list[Chapter] = []
        order = 0

        for item in table_of_contents:
            # Skip non-chapter items (e.g., TOC, copyright pages)
            if not self._is_chapter(item):
                logger.trace(f"Skipping non-chapter item: {item}")
                continue

            # Extract title from content or use fallback
            raw_content = item.get_content().decode("utf-8")
            title = self._extract_chapter_title(raw_content) or f"Chapter {order + 1}"
            logger.trace(f"chapter title: {title}")

            # Skip likely table of contents
            if title == self.metadata.title:
                continue

            content = self._clean_text(raw_content)
            # remove the title from the content
            content = self._remove_title_from_content(content, title)

            # Skip empty chapters
            if not content:
                logger.trace(f"Skipping empty chapter: {title}")
                continue

            chapter = Chapter(title=title, content=content, order=order, id=item.id)
            chapters.append(chapter)
            order += 1

        return chapters

    def _extract_chapters(self) -> list[Chapter]:
        """Extract chapters from the EPUB file.

        Returns:
            List[Chapter]: List of extracted chapters
        """
        chapters = []

        # Get the table of contents
        table_of_contents = self.epub.toc
        if table_of_contents:
            logger.trace(f"raw toc: {table_of_contents}")
            table_of_contents = self._flatten_table_of_contents(table_of_contents)
        else:
            table_of_contents = self.epub.get_items_of_type(ebooklib.ITEM_DOCUMENT)
            logger.trace("no toc, using all documents")

        logger.trace(f"toc: {table_of_contents}")

        # If we detected chapters in one page, process the first document
        if self._chapters_in_one_page:
            chapters = self._get_chapters_in_one_page(
                table_of_contents[0], table_of_contents[1:]
            )
        else:
            chapters = self._get_chapters_from_pages(table_of_contents)

        if not chapters:
            raise ConversionError(
                "No valid chapters found in EPUB file", ErrorCodes.INVALID_EPUB
            )

        if self.metadata.title:
            chapters.insert(0, self._get_title_chapter())
            logger.trace(f"Book title added: {chapters[0]}")

        return sorted(chapters)

    def _is_cover(self, item: epub.EpubItem) -> bool:
        """Determine if an EPUB item is a cover.

        Args:
            item: EPUB item to check
        """
        return re.search(r"cover\.x?html$", item.file_name.lower()) is not None

    def _is_chapter(self, item: epub.EpubItem) -> bool:
        """Determine if an EPUB item is a chapter.

        Args:
            item: EPUB item to check

        Returns:
            bool: True if the item is a chapter
        """
        # Skip common non-chapter files
        skip_patterns = [r"toc\.x?html$", r"copyright\.x?html$", r"cover\.x?html$"]
        skip_id_patterns = [r"pg-(header|footer|toc)$"]

        for pattern in skip_patterns:
            if re.search(pattern, item.file_name.lower()):
                return False

        for pattern in skip_id_patterns:
            if re.search(pattern, item.id.lower()):
                return False

        return True

    @staticmethod
    def _extract_chapter_title(raw_content: str) -> Optional[str]:
        """Extract chapter title from an EPUB item.

        Args:
            raw_content: Raw string of html content to extract title from

        Returns:
            Optional[str]: Extracted title or None
        """
        soup = BeautifulSoup(raw_content, "html.parser")

        # Try to find title in common heading elements
        for heading in soup.find_all(["h1", "h2", "h3"]):
            title = heading.get_text(strip=True)
            if title:
                return title

        return None

    @staticmethod
    def _remove_title_from_content(content: str, title: str) -> str:
        """Remove the title from the content.

        Args:
            content: Content to remove title from
            title: Title to remove

        Returns:
            str: Content with title removed
        """
        title_split = title.split()
        content_split = content[: len(title)].split()
        rest_of_content = content[len(title) :]
        while title_split and content_split and title_split[0] == content_split[0]:
            logger.trace(f"removing title from content: {title_split[0]}")
            title_split.pop(0)
            content_split.pop(0)
        content_split.append(rest_of_content)
        content = " ".join(content_split)
        return content



#!/usr/bin/env python3
"""Command-line interface for EPUB to audiobook conversion."""

import re
import sys
import time
from pathlib import Path
from typing import Union

import click
import roman
from loguru import logger
from soundfile import SoundFile  # type: ignore
from tqdm import tqdm  # type: ignore

from .audio_converter import AudioConverter
from .audio_handler import AudioHandler
from .config import (
    DEFAULT_LOGGER_ID,
    DEFAULT_SPEECH_RATE,
    ErrorCodes,
)
from .epub_processor import Chapter, EpubProcessor, get_book_length
from .helpers import (
    ROMAN_REGEX,
    AudioHandlerError,
    ConversionError,
    StrPath,
    check_disk_space,
    clean_filename,
    ensure_dir_exists,
    format_time,
    get_duration,
)
from .voices import Voice


class Epub2Audio:
    """An epub converted into an audio book with AI."""

    def __init__(
        self,
        epub_path: StrPath,
        output_path: StrPath = "",
        voice: Union[str, Voice] = Voice.AF_HEART,
        speech_rate: float = 1.0,
        quiet: bool = True,
        cache: bool = True,
        convert: bool = True,
        max_chapters: int = -1,
        format: str = "ogg",
    ):
        """Creates an AudioBook from an Epub.

        Args:
            epub_path: Path to the EPUB file
            output_path: Path to the output audio file
            voice: Name of the voice to use
            speech_rate: Speech rate multiplier
            quiet: Whether to suppress progress reporting
            cache: Whether to enable caching of audio segments
            convert: Whether to convert the epub immediately
            max_chapters: Maximum number of chapters to process, or -1 for no limit.
            format: Format to use for the output file.
        """
        self.cache = cache
        self.quiet = quiet
        self.epub_path = epub_path
        self.voice = voice
        self.speech_rate = speech_rate
        self.max_chapters = max_chapters
        self.extension = f".{format}"
        if output_path and Path(output_path).suffix != self.extension:
            raise ValueError(f"Output path must have the extension {self.extension}")
        self._parse_epub()
        # Create output filename
        if not output_path and self.metadata.title:
            self.output_path = Path(
                clean_filename(f"{self.metadata.title}{self.extension}")
            )
        elif not output_path:
            self.output_path = Path(epub_path).with_suffix(self.extension)
        elif Path(output_path).suffix != self.extension:
            self.output_path = Path(output_path).with_suffix(self.extension)
        else:
            self.output_path = Path(output_path)

        # Create output directory
        ensure_dir_exists(self.output_path.parent)

        self.current_audibook_time = 0.0

        self.converter = AudioConverter(
            self.epub_path,
            voice=self.voice,
            speech_rate=self.speech_rate,
            cache=self.cache,
            extension=self.extension,
        )
        self.audio_handler = AudioHandler(
            self.epub_path,
            self.output_path,
            self.metadata,
            self.quiet,
        )

        # Estimate required disk space (rough estimate: 1MB per minute of audio)
        estimated_space = get_book_length(self.chapters) * 100  # Very rough estimate
        check_disk_space(self.output_path.parent, estimated_space)

        if convert:
            self.convert()

    @property
    def title(self) -> str:
        """The title of the book."""
        if not self.metadata:
            return ""
        return self.metadata.title

    def _parse_epub(self) -> None:
        """Parses out info from epub."""
        if not self.quiet:
            logger.info(f"Processing EPUB file: {self.epub_path}")

        # Process EPUB
        self.epub = EpubProcessor(self.epub_path)
        self.metadata = self.epub.metadata
        self.chapters = self.epub.chapters
        self.warnings = self.epub.warnings
        uses_roman_numerals = True
        for chapter in self.chapters[1:]:
            if not re.search(r"^chapter\s+[ivxclm]+\s", chapter.title.lower()):
                uses_roman_numerals = False
                break

        if uses_roman_numerals:
            logger.debug("Using roman numerals for chapter markers")
            self._roman_to_arabic()

    def _roman_to_arabic(self) -> None:
        """Convert roman numerals to arabic numerals."""
        for chapter in self.chapters:
            chapter_number = ROMAN_REGEX.search(chapter.title)
            if chapter_number:
                chapter_number = roman.fromRoman(chapter_number.group("number"))
                chapter.title = re.sub(
                    ROMAN_REGEX, f"Chapter {chapter_number} ", chapter.title
                )

    def _process_epub_chapter(self, chapter: Chapter) -> None:
        if self.max_chapters > 0 and len(self.chapters) > self.max_chapters:
            logger.info(f"Skipping chapter {chapter.title} as max chapters reached")
            return

        # Generate chapter announcement
        announcement = self.converter.convert_text(chapter.title)
        self.audio_segments.append(announcement)
        logger.debug(
            f"Chapter: '{chapter.title}' "
            f"announcement duration: {get_duration(announcement)}"
        )

        # Convert chapter text
        logger.trace(
            f"start converting chapter '{chapter.title}' content: "
            f"{len(chapter.content)}"
        )
        chapter_audio = self.converter.convert_text(chapter.content)
        self.audio_segments.append(chapter_audio)
        logger.debug(
            f"Chapter: '{chapter.title}' audio duration: {get_duration(chapter_audio)}"
        )

        # Add chapter marker
        start_time = self.current_audibook_time
        self.current_audibook_time += get_duration(announcement) + get_duration(
            chapter_audio
        )
        self.audio_handler.add_chapter_marker(
            chapter.title, start_time, self.current_audibook_time
        )

    def convert(self) -> None:
        """Process an EPUB file and convert it to an audiobook.

        Args:
            epub_path: Path to the EPUB file
            output_path: Path to the output  OGG file
            voice: Name of the voice to use
            speech_rate: Speech rate multiplier
            bitrate: Output audio bitrate
            quiet: Whether to suppress progress reporting
            cache: Whether to enable caching of audio segments
        """
        self.generation_start_time = time.time()

        # Process chapters
        self.current_audibook_time = 0.0
        self.audio_segments: list[SoundFile] = []

        with tqdm(
            total=get_book_length(self.chapters),
            desc="Converting chapters",
            disable=self.quiet,
            unit="chars",
        ) as pbar:
            for chapter in self.chapters:
                self._process_epub_chapter(chapter)
                pbar.update(len(chapter.content))

        # Concatenate all audio segments
        if not self.quiet:
            logger.info("Finalizing audio file...")

        self.audio_handler.finalize_audio_file(self.audio_segments)

        # Clean up cache files
        if not self.cache:
            self.converter.cache_dir_manager.cleanup()

        if self.quiet:
            return

        self.generation_time = time.time() - self.generation_start_time

        # Display summary info
        display_generation = format_time(self.generation_time)
        display_duration = format_time(self.audio_handler.total_duration)
        logger.success(f"\nAudiobook created successfully: {self.output_path}")
        logger.info(f"Total time generating: {display_generation}")
        logger.info(f"Total audio duration: {display_duration}")
        logger.info(f"Total chapters: {self.audio_handler.total_chapters}")

        # Display any warnings
        if self.warnings:
            logger.info("\nWarnings during conversion:")
            for warning in self.warnings:
                logger.info(f"- {warning.message}")


def process_epub(
    input_epub: StrPath,
    output: StrPath,
    speech_rate: float,
    voice: Union[str, Voice],
    quiet: bool = True,
    cache: bool = True,
    convert: bool = True,
    max_chapters: int = -1,
    format: str = "flac",
) -> Epub2Audio:
    """Process an EPUB file and convert it to an audiobook.

    Args:
        input_epub: Path to the EPUB file
        output: Path to the output OGG file
        voice: Name of the voice to use
        speech_rate: Speech rate multiplier
        quiet: Whether to suppress progress reporting
        cache: Whether to enable caching of audio segments
        convert: Whether to convert the epub immediately
        max_chapters: Maximum number of chapters to process, or -1 for no limit.
        format: Format to use for the output file.
    """
    return Epub2Audio(
        input_epub,
        output,
        voice=voice,
        speech_rate=speech_rate,
        quiet=quiet,
        cache=cache,
        convert=convert,
        max_chapters=max_chapters,
        format=format,
    )


@click.command()
@click.argument(
    "input_epub",
    type=click.Path(
        exists=True, readable=True, file_okay=True, dir_okay=False, path_type=Path
    ),
)
@click.option(
    "--output",
    "-o",
    type=click.Path(exists=False, writable=True, path_type=Path),
    default=None,
    help="Output path for the audiobook OGG file, default to book title",
)
@click.option(
    "--voice",
    "-v",
    type=Voice,
    default=Voice.AF_HEART,
    help="Voice to use for text-to-speech.",
    show_choices=False,
)
@click.option(
    "--speech-rate",
    "-s",
    type=float,
    default=DEFAULT_SPEECH_RATE,
    help="Speech rate multiplier.",
)
@click.option(
    "--format",
    "-f",
    type=str,
    default="ogg",
    help="Format to use for the output file.",
    show_choices=["ogg", "flac", "mp3"],
)
@click.option("--quiet", "-q", is_flag=True, help="Suppress progress reporting.")
@click.option("--cache", "-c", is_flag=True, help="Enable caching of audio segments.")
@click.option("--verbose", "-v", help="Enable verbose mode.", count=True)
@click.option(
    "--max-chapters",
    "-m",
    type=int,
    help="Maximum number of chapters to process, or -1 for no limit.",
    default=-1,
    show_default=True,
)
@click.version_option()
def main(
    input_epub: Path,
    output: Path,
    voice: Union[str, Voice],
    speech_rate: float,
    quiet: bool,
    cache: bool,
    verbose: int,
    convert: bool = True,
    max_chapters: int = -1,
    format: str = "flac",
) -> None:
    """Convert an EPUB ebook to an audiobook.

    INPUT_EPUB is the path to the EPUB file to convert.
    """
    if quiet:
        logger.remove(DEFAULT_LOGGER_ID)
    if verbose:
        logger.remove(DEFAULT_LOGGER_ID)
        if verbose >= 2:
            logger.add(sys.stderr, level="TRACE")
            logger.debug("Logging level: TRACE")
        elif verbose >= 1:
            logger.add(sys.stderr, level="DEBUG")
            logger.debug("Logging level: DEBUG")
        logger.trace(f"Input EPUB: {input_epub}")
        logger.trace(f"Output path: {output}")
        logger.trace(f"Voice: {voice}")
        logger.trace(f"Speech rate: {speech_rate}")
        logger.trace(f"Quiet: {quiet}")
        logger.trace(f"Cache: {cache}")
        logger.trace(f"Convert: {convert}")
        logger.trace(f"Format: {format}")
        logger.trace(f"Max chapters: {max_chapters}")
    try:
        process_epub(
            input_epub,
            output,
            speech_rate,
            voice,
            quiet,
            cache,
            convert,
            max_chapters,
            format,
        )
    except ConversionError as e:
        logger.exception(e)
        logger.error(f"Conversion error: {e.message}", err=True)
        exit(e.error_code)
    except AudioHandlerError as e:
        logger.exception(e)
        logger.error(f"Audio handler error: {e.message}", err=True)
        exit(e.error_code)
    except Exception as e:
        logger.exception(e)
        logger.error(f"Unexpected error: {str(e)}", err=True)
        exit(ErrorCodes.UNKNOWN_ERROR)


if __name__ == "__main__":
    main()



"""Configuration settings for the EPUB to Audiobook converter."""

import sys
from collections import namedtuple
from os import getenv
from pathlib import Path

from loguru import logger
from mutagen.flac import FLAC
from mutagen.mp3 import MP3
from mutagen.oggopus import OggOpus

# Set up logging
try:
    logger.remove(0)
except Exception:
    pass
DEFAULT_LOGGER_ID = logger.add(sys.stderr, level="INFO")

# Audio settings
SAMPLE_RATE = 24000  # Hz
DEFAULT_BITRATE = "192k"
DEFAULT_SPEECH_RATE = 1.0
AUDIO_FORMAT = "ogg"
AUDIO_CHANNELS = 1  # Mono

# File handling
DEFAULT_OUTPUT = "audiobook.ogg"
CACHE_DIR = Path(getenv("XDG_CACHE_HOME", Path.home() / ".cache")) / "epub2audio"

# Progress reporting
PROGRESS_UPDATE_INTERVAL = 0.5  # seconds

KOKORO_PATHS = {
    "config": getenv("KOKORO_CONFIG_PATH", "packages/kokoro-weights/config.json"),
    "model_weight": getenv(
        "KOKORO_MODEL_WEIGHT_PATH", "packages/kokoro-weights/kokoro-v1_0.pth"
    ),
    "voice_weights": getenv(
        "KOKORO_VOICE_WEIGHTS_PATH", "packages/kokoro-weights/voices"
    ),
}


# Error codes
class ErrorCodes:
    """Error codes for the EPUB to Audiobook converter."""

    SUCCESS = 0
    INVALID_EPUB = 1
    INVALID_VOICE = 2
    FILESYSTEM_ERROR = 3
    DISK_SPACE_ERROR = 4
    UNKNOWN_ERROR = 99


# Warning types
class WarningTypes:
    """Warning types for the EPUB to Audiobook converter."""

    NON_TEXT_ELEMENT = "non_text_element"
    UNSUPPORTED_METADATA = "unsupported_metadata"
    FORMATTING_ISSUE = "formatting_issue"


# Metadata fields to preserve
METADATA_FIELDS = [
    "title",
    "creator",
    "date",
    "identifier",
    "language",
    "publisher",
    "description",
]

AudioFormat = namedtuple("AudioFormat", ["format", "subtype", "file_class"])

SUPPORTED_AUDIO_FORMATS = {
    ".flac": AudioFormat("FLAC", "PCM_16", FLAC),
    ".ogg": AudioFormat("OGG", "OPUS", OggOpus),
    ".mp3": AudioFormat("MP3", "MPEG_LAYER_III", MP3),
}



"""Unit tests for command-line interface."""

import os
from collections.abc import Generator
from pathlib import Path
from unittest.mock import Mock, patch

import pytest
from click.testing import CliRunner

from epub2audio.config import ErrorCodes
from epub2audio.epub2audio import main, process_epub
from epub2audio.helpers import ConversionError
from epub2audio.voices import Voice


@pytest.fixture
def cli_runner() -> CliRunner:
    """Create a Click CLI test runner."""
    return CliRunner()


@pytest.fixture
def mock_process_epub() -> Generator[Mock, None, None]:
    """Mock the process_epub function."""
    # Create a mock that can be called with any arguments
    mock = Mock(return_value=Mock())
    with patch("epub2audio.epub2audio.process_epub", mock):
        yield mock


def test_cli_basic(
    cli_runner: CliRunner, mock_process_epub: Mock, tmp_path: Path
) -> None:
    """Test basic CLI usage with default parameters."""
    input_file = tmp_path / "test.epub"
    input_file.touch()

    result = cli_runner.invoke(main, [str(input_file), "--cache"])

    assert result.exit_code == 0
    mock_process_epub.assert_called_once_with(
        input_file, None, 1.0, Voice.AF_HEART, False, True, True, -1, "ogg"
    )


def test_cli_with_options(
    cli_runner: CliRunner, mock_process_epub: Mock, tmp_path: Path
) -> None:
    """Test CLI with all options specified and verify they are passed correctly."""
    input_file = str(tmp_path / "test.epub")
    output = str(tmp_path / "output.ogg")
    open(input_file, "w").close()  # Create empty file

    # Use direct testing approach
    result = cli_runner.invoke(
        main,
        [
            input_file,
            "--output",
            output,
            "--speech-rate",
            "1.5",
            "--quiet",
            "--cache",
        ],
    )

    assert result.exit_code == 0
    # We don't care about the exact argument types, just check that a call happened
    assert mock_process_epub.called


def test_cli_missing_input(cli_runner: CliRunner) -> None:
    """Test CLI fails gracefully when input file is not provided."""
    result = cli_runner.invoke(main, [])
    assert result.exit_code != 0
    assert "Missing argument" in result.output


def test_cli_invalid_input(cli_runner: CliRunner) -> None:
    """Test CLI fails gracefully when input file does not exist."""
    result = cli_runner.invoke(main, ["nonexistent.epub"])
    assert result.exit_code != 0
    assert "does not exist" in result.output


def test_cli_invalid_rate(cli_runner: CliRunner, tmp_path: Path) -> None:
    """Test CLI fails gracefully when an invalid speech rate is provided."""
    input_file = str(tmp_path / "test.epub")
    open(input_file, "w").close()  # Create empty file

    result = cli_runner.invoke(main, [input_file, "--speech-rate", "invalid"])
    assert result.exit_code != 0
    assert "Invalid value" in result.output


def test_process_epub_error_handling(cli_runner: CliRunner, tmp_path: Path) -> None:
    """Test error handling when process_epub raises a ConversionError."""
    input_file = str(tmp_path / "test.epub")
    open(input_file, "w").close()  # Create empty file

    with patch(
        "epub2audio.epub2audio.process_epub",
        side_effect=ConversionError("Test error", ErrorCodes.INVALID_EPUB),
    ):
        with patch("epub2audio.epub2audio.logger.error") as mock_error:
            result = cli_runner.invoke(main, [input_file])
            assert result.exit_code == ErrorCodes.INVALID_EPUB
            mock_error.assert_any_call("Conversion error: Test error", err=True)


def test_process_epub_unexpected_error(cli_runner: CliRunner, tmp_path: Path) -> None:
    """Test error handling when process_epub raises an unexpected error."""
    input_file = str(tmp_path / "test.epub")
    open(input_file, "w").close()  # Create empty file

    with patch(
        "epub2audio.epub2audio.process_epub", side_effect=Exception("Unexpected error")
    ):
        with patch("epub2audio.epub2audio.logger.error") as mock_error:
            result = cli_runner.invoke(main, [input_file])
            assert result.exit_code == ErrorCodes.UNKNOWN_ERROR
            mock_error.assert_any_call("Unexpected error: Unexpected error", err=True)


@pytest.mark.integration
def test_process_epub_integration(tmp_path: Path) -> None:
    """Integration test for EPUB processing with all components mocked."""
    # Create test files and directories
    input_file = str(tmp_path / "test.epub")
    output = tmp_path / "output.ogg"
    os.makedirs(output.parent, exist_ok=True)
    open(input_file, "w").close()  # Create empty file

    # Mock all the necessary components
    with (
        patch("epub2audio.epub2audio.EPUBProcessor") as mock_processor,
        patch("epub2audio.epub2audio.AudioConverter") as mock_converter,
        patch("epub2audio.epub2audio.AudioHandler") as mock_handler,
    ):
        # Set up mock returns
        mock_processor.return_value.extract_metadata.return_value = Mock()
        mock_processor.return_value.extract_chapters.return_value = [
            Mock(title="Chapter 1", content="Test content")
        ]
        mock_processor.return_value.warnings = []

        mock_converter.return_value.convert_text.return_value = Mock()
        mock_converter.return_value.generate_chapter_announcement.return_value = Mock()

        # Run the process
        process_epub(
            input_file, output, speech_rate=1.0, voice="test_voice", quiet=False
        )

        # Verify the process flow
        mock_processor.assert_called_once()
        mock_processor.return_value.extract_metadata.assert_called_once()
        mock_processor.return_value.extract_chapters.assert_called_once()

        mock_converter.assert_called_once()
        assert mock_converter.return_value.convert_text.called
        assert mock_converter.return_value.generate_chapter_announcement.called

        mock_handler.assert_called_once()
        assert mock_handler.return_value.add_chapter_marker.called
        assert mock_handler.return_value.finalize_audio_file.called



"""Unit tests for audio conversion module."""

from collections.abc import Generator
from pathlib import Path
from unittest.mock import Mock, patch

import numpy as np
import pytest
from soundfile import SoundFile

from epub2audio.audio_converter import AudioConverter
from epub2audio.config import SAMPLE_RATE, ErrorCodes
from epub2audio.helpers import ConversionError
from epub2audio.voices import Voice


@pytest.fixture
def mock_tts() -> Generator[Mock, None, None]:
    """Create a mock TTS engine with default test voice and audio output."""
    with patch("epub2audio.audio_converter.KPipeline") as mock_kpipeline:
        # Create mock KModel
        with patch("epub2audio.audio_converter.KModel") as mock_kmodel:
            # Set up mock model
            _ = mock_kmodel.return_value

            # Set up mock TTS engine
            tts = mock_kpipeline.return_value
            voice = Mock()
            voice.value = {"name": "test_voice"}
            voice.name = "test_voice"
            voice.local_path = None
            tts.get_voice.return_value = voice

            # Patch Path.exists to return True for model weights check
            with patch("pathlib.Path.exists", return_value=True):
                # Create 1 second of silence as default audio output
                tts.convert_text.return_value = np.zeros(SAMPLE_RATE, dtype=np.float32)

                # Set up mock for __call__
                mock_result = Mock()
                mock_result.audio = Mock()
                mock_result.audio.numpy.return_value = np.zeros(
                    SAMPLE_RATE, dtype=np.float32
                )
                mock_result.phonemes = "test phonemes"
                tts.return_value = [mock_result]

                yield tts


def test_audio_converter_init(mock_tts: Mock, tmp_path: Path) -> None:
    """Test AudioConverter initialization with default parameters."""
    # Create a dummy EPUB file
    epub_path = tmp_path / "test.epub"
    epub_path.touch()  # Create an empty file

    # Mock the CacheDirManager to prevent it from actually trying to open/read the file
    with patch("epub2audio.audio_converter.CacheDirManager") as mock_cache_dir:
        mock_cache_dir_instance = Mock()
        mock_cache_dir.return_value = mock_cache_dir_instance

        converter = AudioConverter(epub_path=str(epub_path))
        assert converter is not None
        assert converter.speech_rate == 1.0

        # Verify CacheDirManager was called with the correct path and default extension
        mock_cache_dir.assert_called_once_with(str(epub_path), extension=".flac")


def test_audio_converter_init_invalid_voice(mock_tts: Mock, tmp_path: Path) -> None:
    """Test AudioConverter initialization fails with invalid voice name."""
    epub_path = tmp_path / "test.epub"
    epub_path.touch()  # Create an empty file

    # Mock file open operation
    with patch(
        "builtins.open",
        Mock(return_value=Mock(read=Mock(return_value=b"dummy content"))),
    ):
        # Set the mock to raise an exception on get_voice
        mock_tts.get_voice.side_effect = Exception("Invalid voice")

        # Test with invalid voice name
        with pytest.raises(ConversionError) as exc_info:
            AudioConverter(epub_path=str(epub_path), voice="invalid_voice")
        assert exc_info.value.error_code == ErrorCodes.INVALID_VOICE


def test_get_voice(mock_tts: Mock, tmp_path: Path) -> None:
    """Test voice selection with both valid and invalid voice names."""
    test_voice = Voice.AF_HEART
    test_epub = tmp_path / "test.epub"
    test_epub.touch()

    # Test Voice class handling directly instead of the AudioConverter
    # Test valid voice
    voice = Voice.get_by_name(test_voice.name)
    assert voice.name == test_voice.name

    # Test invalid voice
    with patch.object(
        Voice,
        "get_by_name",
        side_effect=ConversionError("Invalid voice", ErrorCodes.INVALID_VOICE),
    ):
        with pytest.raises(ConversionError) as exc_info:
            Voice.get_by_name("invalid_voice")
        assert exc_info.value.error_code == ErrorCodes.INVALID_VOICE


def test_convert_text(mock_tts: Mock, tmp_path: Path) -> None:
    """Test text to speech conversion produces valid audio output."""
    epub_path = tmp_path / "test.epub"
    epub_path.touch()  # Create an empty file

    # Mock the file open operation in CacheDirManager
    with patch(
        "builtins.open",
        Mock(return_value=Mock(read=Mock(return_value=b"dummy content"))),
    ):
        # Mock the CacheDirManager
        with patch("epub2audio.audio_converter.CacheDirManager") as mock_cache_dir:
            mock_cache_dir_instance = Mock()
            mock_cache_dir_instance.get_file.return_value = str(
                tmp_path / "cached_audio.flac"
            )
            mock_cache_dir.return_value = mock_cache_dir_instance

            # Mock os.path.exists to simulate the cached file doesn't exist
            with patch("os.path.exists", return_value=False):
                # Mock SoundFile creation
                mock_sound_file = Mock(spec=SoundFile)
                mock_sound_file.close = Mock()

                with patch(
                    "epub2audio.audio_converter.SoundFile", return_value=mock_sound_file
                ):
                    # Mock os.rename to avoid actual file operations
                    with patch("os.rename"):
                        converter = AudioConverter(epub_path=str(epub_path))

                        # Set up converter._audio_data_generator method to yield mocks
                        mock_result = Mock()
                        mock_result.audio = Mock()
                        mock_result.audio.numpy.return_value = np.zeros(
                            1000, dtype=np.float32
                        )
                        mock_result.phonemes = "test phonemes"

                        with patch.object(
                            converter,
                            "_audio_data_generator",
                            return_value=[mock_result],
                        ):
                            segment = converter.convert_text("Test text")

                            # Verify the correct calls were made
                            assert segment == mock_sound_file


def test_convert_text_error(mock_tts: Mock, tmp_path: Path) -> None:
    """Test text to speech conversion handles TTS engine errors."""
    epub_path = tmp_path / "test.epub"
    epub_path.touch()  # Create an empty file

    # Mock the file open operation in CacheDirManager
    with patch(
        "builtins.open",
        Mock(return_value=Mock(read=Mock(return_value=b"dummy content"))),
    ):
        # Mock the CacheDirManager
        with patch("epub2audio.audio_converter.CacheDirManager") as mock_cache_dir:
            mock_cache_dir_instance = Mock()
            mock_cache_dir_instance.get_file.return_value = str(
                tmp_path / "cached_audio.flac"
            )
            mock_cache_dir.return_value = mock_cache_dir_instance

            with patch("os.path.exists", return_value=False):
                with patch("epub2audio.audio_converter.SoundFile"):
                    converter = AudioConverter(epub_path=str(epub_path))

                    # Set up the _audio_data_generator to raise an exception
                    with patch.object(
                        converter, "_audio_data_generator"
                    ) as mock_generator:
                        mock_generator.side_effect = Exception("TTS error")

                        with pytest.raises(ConversionError) as exc_info:
                            converter.convert_text("Test text")
                        assert exc_info.value.error_code == ErrorCodes.UNKNOWN_ERROR


def test_chapter_announcement_conversion(mock_tts: Mock, tmp_path: Path) -> None:
    """Test chapter announcement conversion with proper formatting."""
    epub_path = tmp_path / "test.epub"
    epub_path.touch()  # Create an empty file

    # Mock the file open operation in CacheDirManager
    with patch(
        "builtins.open",
        Mock(return_value=Mock(read=Mock(return_value=b"dummy content"))),
    ):
        # Mock the CacheDirManager
        with patch("epub2audio.audio_converter.CacheDirManager") as mock_cache_dir:
            mock_cache_dir_instance = Mock()
            mock_cache_dir_instance.get_file.return_value = str(
                tmp_path / "cached_audio.flac"
            )
            mock_cache_dir.return_value = mock_cache_dir_instance

            # Need to mock convert_text to verify chapter announcement formatting
            with patch("epub2audio.audio_converter.SoundFile"):
                converter = AudioConverter(epub_path=str(epub_path))

                # Mock the convert_text method
                with patch.object(converter, "convert_text") as mock_convert_text:
                    mock_soundfile = Mock(spec=SoundFile)
                    mock_convert_text.return_value = mock_soundfile

                    # Test that chapter announcement is formatted correctly
                    chapter_title = "Chapter 1"
                    converter.convert_text(chapter_title)

                    # Verify the convert_text was called with the chapter title
                    mock_convert_text.assert_called_with(chapter_title)



"""Text-to-speech conversion using Kokoro."""

import os
from collections.abc import Generator
from pathlib import Path
from typing import Union

import numpy as np
from kokoro import KModel, KPipeline  # Maybe I'll implement TextToSpeech, Voice
from loguru import logger
from soundfile import SoundFile

from .config import (
    KOKORO_PATHS,
    SAMPLE_RATE,
    SUPPORTED_AUDIO_FORMATS,
    ErrorCodes,
)
from .helpers import (
    CacheDirManager,
    ConversionError,
    StrPath,
)
from .voices import Voice


class AudioConverter:
    """Class for converting text to speech using Kokoro."""

    def __init__(
        self,
        epub_path: StrPath,
        voice: str | Voice = Voice.AF_HEART,
        speech_rate: float = 1.0,
        cache: bool = True,
        extension: str = ".flac",
    ):
        """Initialize the audio converter.

        Args:
            epub_path: Path to the EPUB file, used to generate a cache directory
            voice: Voice to use
            speech_rate: Speech rate multiplier
            cache: Whether to cache the generated audio
            extension: Extension to use for the output file

        Raises:
            ConversionError: If the voice is invalid or TTS initialization fails
        """
        try:
            self.voice = self._get_voice(voice)
            model = True
            if Path(KOKORO_PATHS["model_weight"]).exists():
                logger.debug(f"model weight found: {KOKORO_PATHS['model_weight']}")
                model = KModel(
                    config=KOKORO_PATHS["config"],
                    model=KOKORO_PATHS["model_weight"],
                    repo_id="hexgrad/Kokoro-82M",
                )
            self.tts = KPipeline(
                lang_code=self.voice.lang_code,
                repo_id="hexgrad/Kokoro-82M",
                model=model,
            )
            self.speech_rate = speech_rate
            self.cache_dir_manager = CacheDirManager(epub_path, extension=extension)
            self.cache = cache
            self.extension = extension
            self.format, self.subtype, _ = SUPPORTED_AUDIO_FORMATS[extension]
        except Exception as e:
            raise ConversionError(
                f"Failed to initialize TextToSpeech: {str(e)}", ErrorCodes.INVALID_VOICE
            ) from e

    def _get_voice(self, voice: Union[str, Voice]) -> Voice:
        """Get a voice by name.

        Args:
            voice: Voice to get

        Returns:
            Voice: The requested voice

        Raises:
            ConversionError: If the voice is invalid
        """
        if isinstance(voice, Voice):
            return voice
        try:
            return Voice.get_by_name(voice)
        except Exception as e:
            valid_voices = ", ".join([v.name for v in Voice.list_voices()])
            raise ConversionError(
                f"Invalid voice '{voice}'. Available voices: {valid_voices}",
                ErrorCodes.INVALID_VOICE,
            ) from e

    def _audio_data_generator(
        self, text: str
    ) -> Generator[KPipeline.Result, None, None]:
        """Generate audio data from text.

        Args:
            text: Text to convert

        Returns:
            Generator[KPipeline.Result, None, None]: Audio data generator
        """
        try:
            voice = self.voice.local_path if self.voice.local_path else self.voice.name
            yield from self.tts(
                text, voice=voice, speed=self.speech_rate, split_pattern=r"\n+"
            )
        except Exception as e:
            raise ConversionError(
                f"Failed to generate audio data: {str(e)}", ErrorCodes.UNKNOWN_ERROR
            ) from e

    def convert_text(self, text: str) -> SoundFile:
        """Convert text to speech.

        Args:
            text: Text to convert

        Returns:
            SoundFile: Converted audio
        """
        try:
            # Create a temporary file
            temp_file = self.cache_dir_manager.get_file(text)

            # If the file exists and caching is enabled, return the cached file
            if os.path.exists(temp_file) and self.cache:
                logger.trace(f"returning cached file: {temp_file}")
                return SoundFile(temp_file)

            if os.path.exists(f"{temp_file}.generating"):
                logger.trace(f"removing generating file: {temp_file}.generating")
                os.remove(f"{temp_file}.generating")

            audio_data = SoundFile(
                f"{temp_file}.generating",
                mode="w",
                samplerate=SAMPLE_RATE,
                channels=1,
                format=self.format,
                subtype=self.subtype,
            )
            int_size = np.int16
            max_int_size = np.iinfo(int_size).max
            # Generate speech
            for result in self._audio_data_generator(text):
                phonemes = result.phonemes
                audio = result.audio
                logger.trace(f"Phonemes: {phonemes}")
                if audio is None:
                    continue
                audio_bytes = (audio.numpy() * max_int_size).astype(int_size)
                audio_data.write(audio_bytes)

            # Close the audio data, and rename the file to the final file
            audio_data.close()
            os.rename(f"{temp_file}.generating", temp_file)
            return SoundFile(temp_file)

        except Exception as e:
            raise ConversionError(
                f"Failed to convert text to speech: {str(e)}", ErrorCodes.UNKNOWN_ERROR
            ) from e
